/**
 * @file
 * TestCompilingAndLinking interface and implementation.
 *
 * @copyright Copyright © 2012–2014 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include "TestCompositionExecution.hh"

#include <Vuo/Vuo.h>
#include <stdlib.h>


/**
 * Tests the compiling and linking of compositions.
 */
class TestCompilingAndLinking : public TestCompositionExecution
{
	Q_OBJECT

	string cachedDylib;
	string cachedIndex;
	VuoFileUtilities::File cachedDylibFile;

public:

	TestCompilingAndLinking()
		: cachedDylibFile(VuoFileUtilities::getCachePath(), "libVuoResources.dylib")
	{
		string cacheDir = VuoFileUtilities::getCachePath();
		cachedDylib = cacheDir + "/libVuoResources.dylib";
		cachedIndex = cacheDir + "/index.txt";
	}

private:

	bool doesCacheGetRecreated(VuoCompiler *compiler)
	{
		unsigned long cachedDylibLastModified = (VuoFileUtilities::fileExists(cachedDylib) ?
													 VuoFileUtilities::getFileLastModifiedInSeconds(cachedDylib) : 0);
		unsigned long cachedIndexLastModified = (VuoFileUtilities::fileExists(cachedIndex) ?
													 VuoFileUtilities::getFileLastModifiedInSeconds(cachedIndex) : 0);

		compiler->hasTriedCachedResources = false;
		delete compiler->sharedEnvironment;
		compiler->sharedEnvironment = NULL;
		compiler->getCachedResources();

		return VuoFileUtilities::fileExists(cachedDylib) &&
				VuoFileUtilities::fileExists(cachedIndex) &&
				VuoFileUtilities::getFileLastModifiedInSeconds(cachedDylib) > cachedDylibLastModified &&
				VuoFileUtilities::getFileLastModifiedInSeconds(cachedIndex) > cachedIndexLastModified;
	}

private slots:

	void testCreationOfCachedResources()
	{
		VuoCompiler compiler;  // Not created with VuoCompositionExecution::initCompiler() — don't want extra module folders.

		string nodeClass = "vuo.test.doNothing.vuonode";
		string nodeDir = QDir::current().absolutePath().toStdString() + "/node-TestCompilingAndLinking";
		string moduleDir = VuoFileUtilities::getUserModulesPath();
		string nodeClassInModuleDir = moduleDir + "/" + nodeClass;

		cachedDylibFile.lockForWriting();
		remove(cachedDylib.c_str());
		remove(cachedIndex.c_str());
		remove(nodeClassInModuleDir.c_str());
		cachedDylibFile.unlock();

		// Cached files do not exist.
		QVERIFY( doesCacheGetRecreated(&compiler) );

		// Cached files are up to date.
		QVERIFY( ! doesCacheGetRecreated(&compiler) );

		// Cached dylib does not exist.
		remove(cachedDylib.c_str());
		QVERIFY( doesCacheGetRecreated(&compiler) );

		// Node class added to Modules folder.
		string escapedNodeClassInNodeDir = "\"" + nodeDir + "/" + nodeClass + "\"";
		string escapedModuleDir = "\"" + moduleDir + "\"";
		system(("/bin/cp " + escapedNodeClassInNodeDir + " " + escapedModuleDir).c_str());
		QVERIFY( doesCacheGetRecreated(&compiler) );

		// Node class updated in Modules folder.
		sleep(2);  // Make sure getFileLastModifiedInSeconds() is larger for the node class than the cache files.
		string escapedNodeClassInModuleDir = "\"" + nodeClassInModuleDir + "\"";
		system(("/usr/bin/touch " + escapedNodeClassInModuleDir).c_str());
		QVERIFY( doesCacheGetRecreated(&compiler) );

		// Node class deleted from Modules folder.
		remove(nodeClassInModuleDir.c_str());
		QVERIFY( doesCacheGetRecreated(&compiler) );
	}

};


QTEST_APPLESS_MAIN(TestCompilingAndLinking)
#include "TestCompilingAndLinking.moc"
